import sys
sys.path.append('../stale_data_detection/')
from extract_table_names import get_table_names, get_raw_urls
import argparse

def main():
    parser = argparse.ArgumentParser(description='Testing extract table names from URL')
    parser.add_argument('-u', '--url', help='URL to the GitHub repo')
    parser.add_argument('-b', '--branch', help='name of the GitHub branch')
    parser.add_argument('-a', '--access_token', help='Access token to authenticate')
    parser.add_argument('-e', '--ext', help='file extension to search for')

    args = parser.parse_args()

    access_token = None
    file_ext = '.hql'

    if args.access_token:
        access_token = args.access_token
    if args.ext:
        file_ext = args.ext
    
    if args.url and args.branch:
        repo_url = args.url
        branch_name = args.branch

        raw_urls = get_raw_urls(repo_url, branch_name, access_token, file_ext)
        for url in raw_urls:
            print(url)
        print('\n')
        tables = get_table_names(repo_url, branch_name, access_token, file_ext)
        for table in tables:
            print(table)

    else:
        print("Invalid arguments")
        
if __name__ == "__main__":
    main()


"""
    Python utility function to detect column with last updated date values from 
    Hive tables 
"""

import pandas as pd
import numpy as np
from datetime import datetime

def get_table_update_status(df):
    """
    Args:
        df (pd.DataFrame)

    Returns:
        update_status: dictionary that contains column names as keys 
                        and latest update dates as values
    """
    update_status = {}
    update_columns = get_datetime_fields(df)

    for col in update_columns:
        # remove DOB column
        if is_dob_col(df, col):
            continue

        lastest_date = df[col].max().strftime('%Y-%m-%d %H:%M:%S')
        # ignore columns that contain future date
        if datetime.now() > lastest_date:
            update_status[col] = df[col].max().strftime('%Y-%m-%d %H:%M:%S')

    return update_status

def get_datetime_fields(df):
    """
    Args:
        df (pd.DataFrame)

    Returns:
        relevant_cols: list of columns that likely contain update date info
    """
    # see if column is convertible to datetime format
    relevant_cols = []
    for col in df.columns:
        if df[col].dtype == 'object' and not (df[col] == '').all():
            try:
                df[col] = pd.to_datetime(df[col])
            except (ValueError, OverflowError):
                pass
    
    # add column that contains datetime or date values
    for col in df.select_dtypes(include=[np.datetime64]):
        if not (df[col].isna().all()):
            dt = df[col]
            # ignore columns that contain timestamp only
            if not (dt.dt.date == pd.Timestamp('now').date()).all():
                relevant_cols.append(col)
    return relevant_cols


def is_dob_col(df, col):
    """
    Summary: Check whether a column in a DataFrame is date of birth column or not
    Args:
        df (pd.DataFrame): Pandas DataFrame
        col (str): column name

    Return: True if it is a dob column, False otherwise
    """

    # find the range of the values
    min_date = df[col].min()
    max_date = df[col].max()

    # check if column contains dates outside of this range
    # if it is, it's unlikely to be a dob column
    if min_date.year < 1930 or max_date.year > pd.Timestamp.today().year:
        return False

    # check the distribution of values
    # DOB are likely to be uniformly distributed accross months and days of the year
    # but not accross year. If the column has that characteristic, it's more likely
    # to be a DOB column
    by_month_day = df[col].dt.strftime("%m-%d").value_counts(normalize=True)
    by_year = df[col].dt.year.value_counts(normalize=True)

    # check if avg % of dates that occur each month-day combination > 0.05
    # or avg % of dates that occur each year is < 10% -> not likely a DOB column
    # mean frequency for month-day is supposed to be low for DOB
    # and higher for year
    if by_month_day.mean() > 0.05 or by_year.mean() < 0.1:
        return False

    



"""
    Function to flag tables for PanelRx that are considered stale
"""
import getpass
import sys
sys.path.insert(0, f'/users/{getpass.getuser()}/boa')
import boa
from boa.utils import query_db
from datetime import datetime
import pandas as pd
import numpy as np
from get_table_update_status import get_table_update_status

CHECK_FIELD = 'transient_lastDdlTime'
CHECK_COLUMN = 'prpt_name'
CHECK_PERIOD = 14

def is_stale_table(table_name: str, check_period=CHECK_PERIOD):
    """
    Summary: determine whether the table is stale

    Args:
        table_name (str): table name in string format
        check_period (int): number of days out-of-date to flag a table as stale. Default 14 days
    Returns:
        Table staleness status with reason
    """
    # get table metadata lastest update date
    table_status = get_table_last_modified_date(table_name)
    
    if not table_status:
        return {'stale': 'table not found/not have access'}

    # get column lastest update date
    columns_status = get_columns_update_date(table_name)
    if len(columns_status)>0:
        last_updated_col = max(columns_status, key=columns_status.get)
        col_last_date = datetime.strptime(columns_status[last_updated_col], '%Y-%m-%d %H:%M:%S')
    else:
        col_last_date = table_status   
        last_updated_col = '' 

    # run staleness check 
    return staleness_check(table_status, col_last_date, last_updated_col, check_period)


def staleness_check(table_status, col_last_date, col_name, check_period):
    """
    Summary: Rule based checking whether a table is stale
    Args:
        table_status (datetime): table metadata last update date
        col_last_date (datetime): column most recent update date
        col_name (str): name of column that contain that most recent update date
        check_period (int): number of days out-of-date to flag a table as stale

    Returns:
        staleness status of table with the reason.
    """
    # define rule for staleness check
    datediff =  col_last_date - table_status
    # compare between column last update date and current date
    # if column has not been update for too long (check_period), flag table as stale
    if datediff.days >= 0 and col_name !='':
        datediff_now = (datetime.now()-col_last_date).days
        if datediff_now > check_period:
            return {'stale':f'column {col_name} has not been update for {datediff_now} days since {col_last_date}'}
        else:
            return {'not stale':f'column {col_name} last update was on {col_last_date}'}
    else:
        # check if metadata is updated but table content is not updated
        # flag table as stale if that happened 
        if datediff.days < -check_period:
            diff = abs(datediff.days)
            return {'stale':f'table column {col_name} has not been updated for {diff} days compared to metadata'} 
        else:
            # compare between metadata last update date and current date
            # if metadata has not been update for too long (check_period), flag table as stale
            datediff_now = (datetime.now()-table_status).days 
            if datediff_now > check_period:
                return {'stale':f'table metadata has not been update for {datediff_now} days since {table_status}'}
            else: 
                return {'not stale':f'table metadata last update on {table_status}'}


def flag_tables(table_list):
    """
    Summary: determine stale tables from list of table name

    Args:
        table_list (List[str]): list of table name

    Returns:
        list of tables that failed any of the rules
    """
    stale_list = {}
    # loop through tables list and check for staleness
    for table_name in table_list:
        table_status = is_stale_table(table_name)
        if 'stale' in table_status:
            stale_list[table_name] = table_status['stale']
    return stale_list

def get_columns_update_date(table_name: str):
    """
    Summary: get columns latest update dates from table name

    Args:
        table_name (str): table name

    Returns:
        dictionary that contains column names as keys 
        and latest update dates as values
    """
    # get columns update date
    try:
        df = get_table_dataframe(table_name)
        return get_table_update_status(df)
    except Exception as e:
        return []
    


def get_table_last_modified_date(table_name: str, check_field = CHECK_FIELD, check_column = CHECK_COLUMN):
    """
    Summary: get table metadata and return latest modified date

    Args:
        table_name (str): table name

    Returns:
        last_modified_date (datetime): table last modified date 
    """

    # query to get table metadata last update date
    sql_query = f"""
    SHOW TBLPROPERTIES {table_name} ('{check_field}')
    """
    try:
        # get the table metadata and extract last updated date
        df = query_db(sql_query, db='hive', use_arrow=False, return_df=True)
        last_modified_date = datetime.fromtimestamp(int(df[check_column][0]))
        return last_modified_date
    except Exception as e:
        print("Table not found")
        return None


def get_table_dataframe(table_name: str):
    """
    Summary: connect to db and get table content given table name

    Args:
        table_name (str): table name

    Returns:
        df (pd.DataFrame): table content as DataFrame
    """
    # construct query to read table content to dataframe
    sql_query = f'''
    SELECT * FROM {table_name}
    ORDER BY RAND()
    LIMIT 10000
    ''' 

    try:
        # use boa to read table into pd.DataFrame
        df = query_db(sql_query, db='hive', use_arrow=False, return_df=True)
        return df
    except Exception as e:
        print("Table not found")
        return None




import re
import requests
from sql_metadata import Parser

def get_raw_urls(repo_url, branch_name , access_token=None, ext='.hql'):
    """
    Args:
        repo_url (str): string url to the GitHub repo
        branch_name (str): specify branch name of the repo (main/master/other?)
        access_token (str, optional): GitHub access token. Defaults to None.
        ext (str): file extension. Defaul: .hql
    Returns:
        raw_urls (list[str]): list of raw url to files in the repo
    """

    api_url, sub_dir = get_repo_api_url(repo_url, branch_name)
    if api_url == '':
        return []

    url_parts = repo_url.strip().split('/')
    owner, repo = url_parts[3], url_parts[4]

    # Define headers for API request in case requiring access token
    headers={"Accept":"application/vnd.github.v3+json"}
    if access_token:
        headers["Authorization"] = f"token {access_token}"

    # send GET request to GitHub API to get the file contents
    response = requests.get(api_url, headers = headers)
    
    # Get JSON response and extract the file content
    if response.ok:
        response_json = response.json()
        raw_urls = []
        for item in response_json['tree']:
            if not item['type']=='blob':
                continue
            if sub_dir not in item['path']:
                continue
            if not item['path'].endswith(ext):
                continue
            raw_urls.append(f"https://{url_parts[2]}/raw/{owner}/{repo}/{branch_name}/{item['path']}")
        return raw_urls
    else:
        print('Failed to fetch the files list. Please check the repo url or include access token')


def get_repo_api_url(repo_url, branch_name):
    """
    Args:
        repo_url (str): string url to the GitHub repo
        branch_name (str): specify branch name of the repo (main/master/other?)

    Returns:
        api_url (str): api url for that repo/branch to send HTTPS request
        sub_dir (str): path to subdirectory if it's a subdirectory of a repo/branch
    """
    # check if the given URL is valid
    url_parts = repo_url.strip().split('/')
    if len(url_parts) < 5:
        print("Not a valid repo URL. Please check the URL")
        return ['','']

    # construct the Github API url for the repo
    owner, repo = url_parts[3], url_parts[4]
    api_url = f'https://{url_parts[2]}/api/v3/repos/{owner}/{repo}/git/trees/{branch_name}?recursive=1'
    
    # construct the path to the subdirectory if included in the url
    sub_dir = ''
    branch_tokens = branch_name.split('/')
    for i in reversed(range(len(url_parts))):
        if url_parts[i]== branch_tokens[-1]:
            sub_dir = '/'.join(url_parts[i+1:])
    
    return api_url, sub_dir

def get_table_names(repo_url, branch_name, access_token=None, ext='.hql'):
    """
    Args:
        repo_url (str): string url to the GitHub repo
        branch_name (str): specify branch name of the repo (main/master/other?)
        access_token (str, optional): GitHub access token. Defaults to None.
        ext (str): file extension. Default: .hql

    Returns:
        table_names (set(str)): set of table names from the given repo
    """

    # get raw urls of all files with given extension on that repo
    raw_urls = get_raw_urls(repo_url, branch_name, access_token, ext)
    table_names = set()

    # loop through each url, read content and extract table names
    for url in raw_urls:
        # get file content from url
        file_content = get_file_content(url, access_token)
        # extract table names from file content
        tables = extract_table_names(file_content)
        if len(tables) > 0:
            table_names.update(tables)
    
    return table_names


def extract_table_names(file_content):
    """
    Args:
        file_content (str): content of a file

    Returns:
        table_names (set(str)): set of table names given file content
    """
    file_content = file_content.lower()
    
    # seearch for the hive variables
    # replace all hive variables with table real names
    hive_vars = re.findall(r'hivevar:(.+?)=', file_content)
    for var in hive_vars:
        if len(var.strip().split(" "))>1:
            continue
        # find the correspond hive variables real names
        var_name = re.findall(f'hivevar:{var}=(.+?);',file_content)
        if len(var_name) > 0:
            var = var.strip()
            pattern = f"${{hivevar:{var}}}".strip()
            # replace hive variables with real names
            file_content = re.sub(re.escape(pattern), var_name[0].strip(), file_content)

    # clean file content, remove newline character
    file_content = re.sub("\n", " ", file_content)
    
    # split content into tokens
    tokens = file_content.split(" ")
    
    queries = get_queries(tokens)
    table_names = set()

    # extract table names using sql_metadata
    for query in queries:
        parser = Parser(query)
        try:
            tables = parser.tables
            for table in tables:
                parser.columns
                if table.split(".")[0] not in parser._table_aliases and len(table.split(" ")) <2:
                    table_names.add(table)
        except Exception as e:
            table_names = set()
       
    return table_names

def get_file_content(file_raw_url, access_token=None):
    """
    Args:
        file_raw_url (str): GitHub raw url to file

    Returns:
        content: file content in str
    """

    # construct headers for GET request
    headers={"Accept":"application/vnd.github.v3+json"}
    if access_token:
        headers["Authorization"] = f"token {access_token}"

    try:
        # sending GET request to get the file content
        response = requests.get(file_raw_url, headers=headers)
        if response.status_code == 200:
            content = response.content.decode("utf-8")
            return content
    except requests.exceptions.RequestException:
        print("Error fetching file, check the file raw url or include access token")

def get_queries(tokens):
    """
    Args:
        tokens (list(str)): list of string

    Returns:
        queries (list(str)): list of queries (SELECT or WITH queries) from tokens
    """

    cur_query = []
    queries = []
    on_query = False
    # loop through tokens and identify SELECT or WITH-AS SELECT queries
    for i in range(len(tokens)):
        if on_query:
            # add string to current query, stop when seeing ';' character
            cur_query.append(tokens[i])
            if tokens[i]!='' and tokens[i][-1]==';':
                on_query = False
                query = " ".join(cur_query)
                queries.append(query)
                cur_query = []
        else:
            # check if current query is WITH - AS query
            if tokens[i] == 'with' and tokens[i+2] == 'as' and (i==0 or len(tokens[i-1]) < 1 or tokens[i-1][-1] != '-'):
                cur_query.append(tokens[i])
                on_query = True
            # check if current query is SELECT query
            elif tokens[i] == 'select' and (i==0 or len(tokens[i-1]) < 1 or tokens[i-1][-1] != '-'):
                cur_query.append(tokens[i])
                on_query = True
            else:
                continue
    return queries





from stale_data_detection.extract_table_names import get_table_names, get_raw_urls,get_repo_api_url,extract_table_names,get_file_content,get_queries
from stale_data_detection.flag_stale_table import flag_tables, is_stale_table, staleness_check, get_columns_update_date, get_table_last_modified_date, get_table_dataframe
from stale_data_detection.get_table_update_status import get_table_update_status, get_datetime_fields, is_dob_col

__all__ = ['get_table_names','get_raw_urls','get_queries',  \
            'get_repo_api_url','extract_table_names','get_file_content',\
            'flag_tables','is_stale_table', 'staleness_check', 'get_columns_update_date',\
            'get_table_last_modified_date','get_table_dataframe',\
            'get_table_update_status','get_datetime_fields','is_dob_col']


import sys
sys.path.append('../stale_data_detection/')

import json
from flag_stale_table import is_stale_table
from extract_table_names import get_table_names
import argparse

def main():
    parser = argparse.ArgumentParser(description='Testing extract table names from URL')
    parser.add_argument('-u', '--url', help='URL to the GitHub repo')
    parser.add_argument('-b', '--branch', help='name of the GitHub branch')
    parser.add_argument('-a', '--access_token', help='Access token to authenticate')
    parser.add_argument('-e', '--ext', help='file extension to search for')
    parser.add_argument('-o', '--output', help='path to output file')

    args = parser.parse_args()

    access_token = None
    file_ext = '.hql'
    output_file = 'stale_list.json'

    if args.access_token:
        access_token = args.access_token
    if args.ext:
        file_ext = args.ext
    if args.output:
        output_file = args.output

    if args.url and args.branch:
        repo_url = args.url
        branch_name = args.branch

        tables = get_table_names(repo_url, branch_name, access_token, file_ext)
        
        staleness = {}
        for table_name in tables:
            table_name = table_name.strip()
            print(f'Processing Table {table_name}')
            staleness[table_name] = is_stale_table(table_name)

        with open(output_file, 'w') as output:
            json.dump(staleness, output, indent=4)

    else:
        print("Invalid arguments")


if __name__ == "__main__":
    main()


from setuptools import setup, find_packages
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='stale_data_detection',
    version='0.4',
    author='HaiLe',
    author_email='hai.le@csvhealth.com',
    license='MIT',
    python_requires='>=3.6',
    packages=find_packages(),
    # specify dependency packages
    install_requires = [
        'sql-metadata', 
    ],
    long_description=long_description,
    long_description_content_type='text/markdown',
    description='Python package to extract table names from github url and check whether the tables are stale',
)